// ignore_for_file: avoid_dynamic_calls

import 'dart:async';
import 'dart:isolate';

import 'package:logger/logger.dart';

void _isolateFunction(IsolateMessage message) {
  void _send(dynamic packet) {
    message.sendPort.send(packet);
  }

  final backendReceivePort = ReceivePort();

  message.sendPort.send(backendReceivePort.sendPort);
  backendReceivePort.listen(
    (dynamic message) {
      if (message is Function) {
        if (message is Future Function()) {
          message.call().then(_send);
        }
        _send(message.call());
      }
    },
  );
}

/// {@template backyard}
class IsolateMessage {
  /// {@macro backyard}
  IsolateMessage(this.sendPort);

  /// {@macro backyard}
  final SendPort sendPort;
}

/// {@template backyard}
class Backyard {
  /// {@macro backyard}
  Backyard({bool debugLog = true}) {
    _l = Logger(
      filter: debugLog ? DevelopmentFilter() : _SilenceFilter(),
      output: ConsoleOutput(),
      level: Level.info,
    );
  }

  late Logger _l;

  final _broadcastReceivePort = StreamController<dynamic>.broadcast();
  StreamSubscription? _broadcastReceivePortSubscription;

  final _frontendReceivePort = ReceivePort();
  late SendPort _sendPort;
  late Isolate _isolate;

  /// initialise the backyard
  Future init() async {
    try {
      // pipe events from receive port to broadcast stream
      _frontendReceivePort.listen(_broadcastReceivePort.sink.add);

      _l.i('SPAWINING ISOLATE ...');
      // create isolate with input object and main running function
      _isolate = await Isolate.spawn<IsolateMessage>(
        _isolateFunction,
        IsolateMessage(_frontendReceivePort.sendPort),
        debugName: 'by-isolate',
      );
      _l.i('ISOLATE SPAWNED!');

      _sendPort = await _broadcastReceivePort.stream
          .firstWhere((dynamic message) => message is SendPort) as SendPort;
      _l.i('SENDPORT RECEIVED FROM BACKEND!');
    } catch (e) {
      _l.e(e);
    }
  }

  /// execute an arbitrary function on the backyard isolate
  Future<RT> execute<RT>(RT Function() function) async {
    try {
      final timer = Stopwatch()..start();
      final result = await _awaitSend<RT>(
        _broadcastReceivePort.stream,
        _sendPort,
        function,
      );
      timer.stop();
      _l.i(
        'EXECUTION TOOK ${timer.elapsedMilliseconds}ms with result: $result',
      );

      return result;
    } catch (e) {
      _l.e(e);
      rethrow;
    }
  }

  Future<RT> _awaitSend<RT>(
    Stream _receivePort,
    SendPort _sendPort,
    Object _sendObject,
  ) async {
    _sendPort.send(_sendObject);
    return await _receivePort.firstWhere((dynamic element) => element is RT)
        as RT;
  }

  /// release all resources attached to the backyard and kill the isolate
  void cancel() {
    _isolate.kill();
    _broadcastReceivePortSubscription?.cancel();
  }
}

class _SilenceFilter extends LogFilter {
  @override
  bool shouldLog(LogEvent event) {
    return false;
  }
}
